(********************************************************************
The grammar for the language you will implement is given below:


Z ::=  ... -1 | 0 | 1 | ...
B ::=  true | false
p ::=  > | < | = | && | || | * | + | -
f,x,y,.. \in Var

e ::=  Z | B | x 
       | p(e1,...en)
       | let x = e in e
       | rec f(e1,...,xn) -> e in e
       | e(e1,...,en)
       | if e then e else e
       | (e)
*********************************************************************)


module rec Exp :
sig
  type var = string 
  type funName = string

  type name = Var of var | Fun of funName

  type valueKind = Fun | Var | Rec

  type prim = Plus | Minus | Times | Equal | Lt | Gt | And | Or 

  type env = var -> value and

  value = I of int | B of bool | Fn of (env * var list * t) | Rec of (funName * env * var list * t)

  and aexp =
      Num of int
    | Bool of bool
    | Lam of (var list * exp)
    | AVar of var

  and cexp =
    | APrimCall of (prim * aexp list)
    | ACall of (aexp * aexp list)
    | AIf of (aexp * exp * exp)

  and exp =
    | ABind of (var * cexp * exp)
    | ABindSimp of (var * aexp * exp)
    | ABindRec of (funName * var list * exp * exp)
    | Cexp of cexp
    | Aexp of aexp
        
  and t =
      Const of value
    | Fun of (var list * t)
    | Var of var
    | PrimCall of (prim * t list)
    | Bind of (var * t * t)
    | BindRec of (funName * var list * t * t)
    | Call of (t * t list)
    | If of (t * t * t) 

  type cont = exp -> exp

  type controlCont = t -> t
  type control = (env * controlCont) list

  val emptyEnv : var -> value
  val update : env -> var -> value -> env
  val normalize : t -> cont -> exp
  val normalizeTerm : t -> exp
  val aNormalToTerm : exp -> t
  val interp : t -> value
end =
struct

  exception IllTypedExpr of string
  exception BindingNotFound of string
  exception ANormalTranslationError of string
  exception UnboundName of string
  exception SmallStepError of string


  type var = string 
  type funName = string 

  type name = Var of var | Fun of funName

  type valueKind = Fun | Var | Rec

  type prim = Plus | Minus | Times | Equal | Lt | Gt | And | Or

  type env = var -> value and

  value = I of int | B of bool |  Fn of (env * var list * t) | Rec of (funName * env * var list * t)

  (* The language of A-normal forms *)
  and aexp =
      Num of int
    | Bool of bool
    | Lam of (var list * exp)
    | AVar of var

    and cexp =
    | APrimCall of (prim * aexp list)
    | ACall of (aexp * aexp list)
    | AIf of (aexp * exp * exp)

    and exp =
    | ABind of (var * cexp * exp)
    | ABindSimp of (var * aexp * exp)
    | ABindRec of (funName * var list * exp * exp)
    | Cexp of cexp
    | Aexp of aexp

  (* The language of expressions *)
  and t =
      Const of value
    | Fun of (var list * t)
    | Var of var
    | PrimCall of (prim * t list)
    | Bind of (var * t * t)
    | BindRec of (funName * var list * t * t)
    | Call of (t * t list)
    | If of (t * t * t) 

  (* Continuations are used to record next steps in the A-normal translation *)
  type cont = exp -> exp
  type conts = exp list -> exp    
  let basecont = fun (e: exp) -> e
    

  (* Environments used in smallstep evaluation *)
  let emptyEnv = fun x -> raise (UnboundName x)
  let update env x v = fun var -> if (var = x) then v else (env var)
                    

  (* Converting an A-Normal expression to a normal expression *)
  let rec aExpNormalToTerm =
    function
    | Num i -> Const (I i)
    | Bool b -> Const (B b)
    | Lam (args,body) -> Fun (args,aNormalToTerm body)
    | AVar x -> Var x
  and
  cExpNormalToTerm =
    function
    | APrimCall(p,args) ->  PrimCall(p,List.map aExpNormalToTerm args)
    | ACall (f,args) -> Call(aExpNormalToTerm f,List.map aExpNormalToTerm args)
    | AIf(guard,t,f) -> If(aExpNormalToTerm guard,aNormalToTerm t, aNormalToTerm f)
  and
  aNormalToTerm =
    function 
    | Aexp e -> aExpNormalToTerm e
    | Cexp c -> cExpNormalToTerm c
    | ABind(x,b,e) -> Bind(x,cExpNormalToTerm b,aNormalToTerm e)
    | ABindSimp(x,b,e) -> Bind(x,aExpNormalToTerm b,aNormalToTerm e)                        
    | ABindRec(f,args,body,e) -> BindRec(f,args,aNormalToTerm body,aNormalToTerm e)


  let genSym =
    let c = ref 0
    in fun () -> c := !c + 1; "G" ^ (string_of_int !c)


  let isValue (e : exp) =
    match e with
       | (Aexp (AVar _)) -> true
       | (Aexp (Num _)) -> true
       | (Aexp (Bool _)) -> true
       | (Aexp (Lam _)) -> true
       | _ -> false

  (* Convert a simple expression to A-normal form *)
  let simpToAexp =
    (function
      | Const v ->
        (match v with
         | I i -> (Num i)
         | B b -> (Bool b)
         | _ -> raise (ANormalTranslationError("in simpToAexp")))
      | Var x -> AVar x
      | _ -> raise (ANormalTranslationError("in simpToAexp")))



  
  (* Introduce a let-binding for a complex expression *) 
  let rec normalizeName (e : t) (k : cont)  =
    normalize e
      (fun e' ->
         if (isValue e')
         then (k e')
         else let t = genSym() in
           match e' with
           | Cexp e'' -> (ABind (t, e'', (k (Aexp (AVar t)))))
           | _ -> Print.printExp e; raise (ANormalTranslationError "normalizeName"))
  and
    (* Generalize to a list of names *)
    normalizeNames (es : t list) (k : conts) =
    (match es with
     | [] -> k []
     | e :: es' -> normalizeName e
                     (fun t -> normalizeNames es' (fun ts -> k (t :: ts))))
  and normalizeTerm (e : t): exp  = normalize e basecont
  and
    (* Given an expression in the term language (t), return its A-normal equivalent *)
    (* You might find the explanation given in: https://matt.might.net/articles/a-normalization/ *)
    (* helpful.  The original paper describing the technique can be found at: https://dl.acm.org/doi/10.1145/173262.155113 *)
    
    normalize (e: t) (k : cont) : exp =
    match e with
    | Fun (vars, body) -> k (Aexp (Lam (vars, normalizeTerm body)))
    | PrimCall (f, args) ->
      normalizeNames args
        (fun ts -> 
          (k (Cexp (APrimCall(f,
                              List.map
                                (fun t ->
                                   match t with
                                   | Aexp aexp -> aexp
                                   | _ -> raise (ANormalTranslationError("Primcall")))
                                ts)))))
    | Const _ -> k (Aexp (simpToAexp e))
    | Var _ -> k (Aexp (simpToAexp e))
    (* FILL-IN HERE *)        



  (* The second part of this exercise is to implement a smallstep interpreter for the
     term language.  The interpreter (smallstep) operates over an expression of type
     t and a control stack (control).  Every entry in the stack contains an
     environment and a continuation representing the bindings the current expression
     can reference and the continuation it should invoke when it is done.  Unlike the
     interpreters you've written before, each time a value is computed, the interpreter
     needs to adjust the control-stack and invoke the continuation to resume the remaining
     part of the computation.  In other words, given a complex expression e, the smallstep
     evaluator decomposes e into successivey smaller components, threading evaluation of
     these components through continuations it builds.  This continuation building is
     particularly apparent in let-expressions where the binding-expression may be complex,
     and conditionals.  Environments maintained by the control-stack are manipulated during
     function calls to handle bindings for the arguments.
  *)      

  
  type controlCont = t -> t
  type control = (env * controlCont) list
      
  let rec smallstep (e : t) (k : control)  =
    match k with
    | [] ->
      (match e with
       | Const v -> v
       | _ -> Print.printExp e; raise (SmallStepError "empty control stack"))
    | (env,k') :: rest ->
      (match e with
       | Const v -> smallstep (k' e) rest
       | Var x ->  smallstep (k' (Const (env x))) rest
       | Fun (args,body) -> smallstep (k' (Const (Fn (env, args, body)))) rest
       (* FILL-IN HERE *))

  (* A-normalize and then un-Anormalize an expression and return the result of applying
     the smallstep evaluator *)
  let interp e =
    let n = normalizeTerm e in
    let ae = aNormalToTerm n in
    (smallstep ae [(emptyEnv, fun v -> v)])
end


and Print :
sig 
  val printExp : Exp.t -> unit
end =
struct

  open Exp
  open Format
      
  let const_int = print_int
  let const_bool = print_bool    
  let kwd = print_string
  let ident = print_string

  let prim = function
    | Plus -> kwd " + "
    | Minus -> kwd " - "
    | Times -> kwd " * "
    | Equal -> kwd " = "
    | Lt -> kwd " < "
    | Gt -> kwd " > "
    | And -> kwd " && "
    | Or -> kwd " ||  "

  let printParams args =
    (kwd "(";
     let rec loop l =
       match l with
       | [] -> kwd ") "
       | [x] -> ident x; kwd ") "
       | x :: l' -> ident x; kwd ","; loop l'
     in loop args)

  let rec printArgs args =
    (kwd "(";
     let rec loop l =
       match l with
       | [] -> kwd ")"
       | [x] -> printExp x; kwd ") "
       | x :: l' -> printExp x; kwd ","; loop l'
     in loop args)
  and printExp = function
    | Const c ->
      (match c with
        | I n -> const_int n
        | B b -> const_bool b
        | Fn (_,args,t) -> (kwd "fn"; printParams args; kwd "-> "; printExp t)
        | Rec (f,_,args,t) -> (kwd "rec "; ident f; printParams args; kwd "= "; printExp t))
    | Fun(args,t) -> (kwd "fn"; printParams args; kwd "-> "; printExp t)
    | Var x -> ident x                    
    | PrimCall(p, args) ->
       kwd "("; printExp (List.nth args 0); prim p; printExp (List.nth args 1); kwd ")"
    | Bind(x,e1,e2) ->
      open_vbox 1; kwd " let "; ident x; kwd "= "; printExp e1; print_break 0 0; kwd "in "; printExp e2; close_box()
    | BindRec(f,args,body,e) ->
      open_vbox 1; kwd " rec "; ident f;  printParams args; kwd "-> "; printExp body; print_break 0 0; kwd "in "; printExp e; close_box()        
    | Call (t, args) ->  open_vbox 2; printExp t;  printArgs args; close_box ()
    | If(e1,e2,e3) ->
      open_vbox 2;  kwd "if "; printExp e1; print_break 1 2; 
      kwd " then "; printExp e2; print_break 1 2;
      kwd " else "; printExp e3; close_box ()

end      

   
(*

let ex1 = (parse "let x = 1 in x + 1")
normalizeTerm ex1 => let x = 1 in x + 1
interp ex1 => 2

let ex2 = (parse "let x = (1 + 2) * (3 + 4) in x + 1")
normalizeTerm ex2 => let G1 = 1 + 2 in let G2 = 3 + 4 in let x = G1 * G2 in x + 1
interp ex2 => 22

let ex3 = (parse "let f = fun(x) -> let z = x + 1 in z in let y = f(2) + f(3) in y + 1")
normalizeTerm ex3 => let f = fun(x) -> let z = x + 1 in z in let G1 = f(2) in G2 = f(3) in y = G1 + G2 in y + 1
interp ex3 => 8

let ex4 = (parse "let x = 1 in let y = 2 in x + y") 
normalizeTerm ex4 => let x = 1 in let y = 2 in x + y
interp ex4 =>  3

let ex5 = (parse "let x = 1 in let f = fun(y) -> x + y in f(3)") 
normalizeTerm ex5 => let x = 1 in let f = fun(y) -> x + y in f(3)") 
interp ex5 => 4

let ex6 = (parse "let f = fun(x,y) -> x + y in let b = true in f(10,  if b then 20 else 30)")
normalizeTerm ex6 => lt f = fun(x,y) -> x + y in let b = true in let G1 = if b then 20 else 30 in f(10,G1)
interp ex6 => 30

let ex7 = (parse "let f = fun (x) -> let g = fun (x) -> x + 1 in g(2) in f(3)") 
normalizeTerm ex7 => let f = fun (x) -> let g = fun (x) -> x + 1 in g(2) in f(3)
interp ex6 -> 3

let ex8 = (parse "rec fact(n) -> if (n = 0) then 1 else n * (fact (n - 1)) in let f = fun(g) -> g(5) in f(fact)");; 
normalizeTerm ex8 => rec fact(n) -> let G1 = (n = 0) in if G1 = 1 then let G2 = n - 1 in let G3 = fact(G2) in n * G3 in let f = fun(g) -> g(5) in f(fact)
interp ex8 => 120

let ex9 = (parse "rec factCps(n,k) -> if (n = 0) then k(1) else factCps(n - 1, fun (v) -> k( n * v)) in factCps(5, fun (x) -> x)");;
normalizeTerm ex9 =>
   rec factCps(n,k) ->
     let G1 = (n = 0) in
     if G1
        then k(1)
        else let G2 = n - 1 in
             factCps(G2, (fun (v) -> let G3 = n * v in k(G3)))
   in factCps(5, fun (x) -> x)
      
interp ex9 => 120

let ex10 =  (parse "let f = fun(x) -> fun(y) -> x + y in let g = fun(h) -> h(2) in g(f(3))");;
normalizeTerm ex10 => let f = fun(x) -> fun(y) -> x + y in let g = fun(h) -> h(2) in let G1 = f(3) in g(G1)
interp ex10 => 5

*)
